import hmac
import hashlib
import aiohttp
import asyncio
import orjson
from typing import Any, Dict, List
from urllib.parse import urljoin, urlencode, unquote

from nexusbridge.base import ApiClient
from nexusbridge.bybit.constants import BybitUrl
from nexusbridge.bybit.error import BybitError


class BybitApiClient(ApiClient):
    def __init__(
        self,
        api_key: str = None,
        secret: str = None,
        url: BybitUrl = BybitUrl.TESTNET,
        timeout: int = 10,
        max_rate: int | None = None,
    ):
        """
        ### Testnet:
        `https://api-testnet.bybit.com`

        ### Mainnet:
        (both endpoints are available):
        `https://api.bybit.com`
        `https://api.bytick.com`

        ### Important:
        Netherland users: use `https://api.bybit.nl` for mainnet
        Hong Kong users: use `https://api.byhkbit.com` for mainnet
        Turkey users: use `https://api.bybit-tr.com` for mainnet
        Kazakhstan users: use `https://api.bybit.kz` for mainnet
        """

        super().__init__(
            api_key=api_key,
            secret=secret,
            timeout=timeout,
            max_rate=max_rate,
        )
        self._recv_window = 5000

        self._base_url = url.base_url

        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

        if api_key:
            self._headers["X-BAPI-API-KEY"] = api_key

    def _generate_signature(self, payload: str) -> List[str]:
        timestamp = str(self._clock.timestamp_ms())

        param = str(timestamp) + self._api_key + str(self._recv_window) + payload
        hash = hmac.new(
            bytes(self._secret, "utf-8"), param.encode("utf-8"), hashlib.sha256
        )
        signature = hash.hexdigest()
        return [signature, timestamp]

    async def _fetch(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        if self._limiter:
            await self._limiter.wait()

        self._init_session()

        url = urljoin(self._base_url, endpoint)
        payload = payload or {}

        payload_str = (
            unquote(urlencode(payload))
            if method == "GET"
            else orjson.dumps(payload).decode("utf-8")
        )

        headers = self._headers
        if signed:
            signature, timestamp = self._generate_signature(payload_str)
            headers = {
                **headers,
                "X-BAPI-TIMESTAMP": timestamp,
                "X-BAPI-SIGN": signature,
                "X-BAPI-RECV-WINDOW": str(self._recv_window),
            }

        if method == "GET":
            url += f"?{payload_str}"
            payload_str = None

        try:
            self._log.debug(f"Request: {url} {payload_str}")
            response = await self._session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_str,
            )
            raw = await response.read()
            if response.status >= 400:
                raise BybitError(
                    code=response.status,
                    message=orjson.loads(raw) if raw else None,
                )
            bybit_response = orjson.loads(raw)
            if bybit_response["retCode"] == 0:
                return raw
            else:
                raise BybitError(
                    code=bybit_response["retCode"],
                    message=bybit_response["retMsg"],
                )
        except aiohttp.ClientError as e:
            self._log.error(f"Client Error {method} Url: {url} {e}")
            raise
        except asyncio.TimeoutError:
            self._log.error(f"Timeout {method} Url: {url}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} {e}")
            raise

    # Account
    from nexusbridge.bybit.api.account import get_v5_account_wallet_balance
    from nexusbridge.bybit.api.account import get_v5_account_withdrawal
    from nexusbridge.bybit.api.account import post_v5_account_upgrade_to_uta
    from nexusbridge.bybit.api.account import get_v5_account_borrow_history
    from nexusbridge.bybit.api.account import post_v5_account_quick_repayment
    from nexusbridge.bybit.api.account import post_v5_account_set_collateral_switch
    from nexusbridge.bybit.api.account import (
        post_v5_account_set_collateral_switch_batch,
    )
    from nexusbridge.bybit.api.account import get_v5_account_collateral_info
    from nexusbridge.bybit.api.account import get_v5_asset_coin_greeks
    from nexusbridge.bybit.api.account import get_v5_account_fee_rate
    from nexusbridge.bybit.api.account import get_v5_account_info
    from nexusbridge.bybit.api.account import get_v5_account_query_dcp_info
    from nexusbridge.bybit.api.account import get_v5_account_transaction_log
    from nexusbridge.bybit.api.account import get_v5_account_contract_transaction_log
    from nexusbridge.bybit.api.account import get_v5_account_smp_group
    from nexusbridge.bybit.api.account import post_v5_account_set_margin_mode
    from nexusbridge.bybit.api.account import post_v5_account_set_hedging_mode
    from nexusbridge.bybit.api.account import post_v5_account_mmp_modify
    from nexusbridge.bybit.api.account import post_v5_account_mmp_reset
    from nexusbridge.bybit.api.account import get_v5_account_mmp_state

    # Asset
    from nexusbridge.bybit.api.asset import get_v5_asset_delivery_record
    from nexusbridge.bybit.api.asset import get_v5_asset_settlement_record
    from nexusbridge.bybit.api.asset import get_v5_asset_exchange_order_record
    from nexusbridge.bybit.api.asset import get_v5_asset_coin_query_info
    from nexusbridge.bybit.api.asset import get_v5_asset_transfer_query_sub_member_list
    from nexusbridge.bybit.api.asset import get_v5_asset_transfer_query_asset_info
    from nexusbridge.bybit.api.asset import (
        get_v5_asset_transfer_query_account_coin_balance,
    )
    from nexusbridge.bybit.api.asset import get_v5_asset_withdraw_withdrawable_amount
    from nexusbridge.bybit.api.asset import (
        get_v5_asset_transfer_query_transfer_coin_list,
    )
    from nexusbridge.bybit.api.asset import post_v5_asset_transfer_inter_transfer
    from nexusbridge.bybit.api.asset import (
        get_v5_asset_transfer_query_inter_transfer_list,
    )
    from nexusbridge.bybit.api.asset import post_v5_asset_transfer_universal_transfer
    from nexusbridge.bybit.api.asset import (
        get_v5_asset_transfer_query_universal_transfer_list,
    )
    from nexusbridge.bybit.api.asset import get_v5_asset_deposit_query_allowed_list
    from nexusbridge.bybit.api.asset import post_v5_asset_deposit_to_account
    from nexusbridge.bybit.api.asset import get_v5_asset_deposit_query_record
    from nexusbridge.bybit.api.asset import get_v5_asset_deposit_query_sub_member_record
    from nexusbridge.bybit.api.asset import get_v5_asset_deposit_query_internal_record
    from nexusbridge.bybit.api.asset import get_v5_asset_deposit_query_address
    from nexusbridge.bybit.api.asset import (
        get_v5_asset_deposit_query_sub_member_address,
    )
    from nexusbridge.bybit.api.asset import get_v5_asset_withdraw_query_record
    from nexusbridge.bybit.api.asset import get_v5_asset_withdraw_query_vasp_list
    from nexusbridge.bybit.api.asset import post_v5_asset_withdraw_create
    from nexusbridge.bybit.api.asset import post_v5_asset_withdraw_cancel
    from nexusbridge.bybit.api.asset import get_v5_asset_exchange_query_coin_list
    from nexusbridge.bybit.api.asset import post_v5_asset_exchange_quote_apply
    from nexusbridge.bybit.api.asset import post_v5_asset_exchange_convert_execute
    from nexusbridge.bybit.api.asset import get_v5_asset_exchange_convert_result_query
    from nexusbridge.bybit.api.asset import get_v5_asset_exchange_query_convert_history

    # Market
    from nexusbridge.bybit.api.market import get_v5_market_time
    from nexusbridge.bybit.api.market import get_v5_market_kline
    from nexusbridge.bybit.api.market import get_v5_market_mark_price_kline
    from nexusbridge.bybit.api.market import get_v5_market_index_price_kline
    from nexusbridge.bybit.api.market import get_v5_market_premium_index_price_kline
    from nexusbridge.bybit.api.market import get_v5_market_instruments_info
    from nexusbridge.bybit.api.market import get_v5_market_orderbook
    from nexusbridge.bybit.api.market import get_v5_market_tickers
    from nexusbridge.bybit.api.market import get_v5_market_funding_history
    from nexusbridge.bybit.api.market import get_v5_market_recent_trade
    from nexusbridge.bybit.api.market import get_v5_market_open_interest
    from nexusbridge.bybit.api.market import get_v5_market_historical_volatility
    from nexusbridge.bybit.api.market import get_v5_market_insurance
    from nexusbridge.bybit.api.market import get_v5_market_risk_limit
    from nexusbridge.bybit.api.market import get_v5_market_delivery_price
    from nexusbridge.bybit.api.market import get_v5_market_account_ratio

    # Position
    from nexusbridge.bybit.api.position import get_v5_position_list
    from nexusbridge.bybit.api.position import post_v5_position_set_leverage
    from nexusbridge.bybit.api.position import post_v5_position_switch_isolated
    from nexusbridge.bybit.api.position import post_v5_position_switch_mode
    from nexusbridge.bybit.api.position import post_v5_position_trading_stop
    from nexusbridge.bybit.api.position import post_v5_position_set_auto_add_margin
    from nexusbridge.bybit.api.position import post_v5_position_add_margin
    from nexusbridge.bybit.api.position import get_v5_position_closed_pnl
    from nexusbridge.bybit.api.position import post_v5_position_move_positions
    from nexusbridge.bybit.api.position import post_v5_position_confirm_pending_mmr
    from nexusbridge.bybit.api.position import post_v5_position_set_tpsl_mode
    from nexusbridge.bybit.api.position import post_v5_position_set_risk_limit
    from nexusbridge.bybit.api.position import get_v5_position_move_history

    # Trade
    from nexusbridge.bybit.api.trade import post_v5_order_create
    from nexusbridge.bybit.api.trade import post_v5_order_amend
    from nexusbridge.bybit.api.trade import post_v5_order_cancel
    from nexusbridge.bybit.api.trade import get_v5_order_realtime
    from nexusbridge.bybit.api.trade import post_v5_order_cancel_all
    from nexusbridge.bybit.api.trade import get_v5_order_history
    from nexusbridge.bybit.api.trade import get_v5_execution_list
    from nexusbridge.bybit.api.trade import post_v5_order_create_batch
    from nexusbridge.bybit.api.trade import post_v5_order_amend_batch
    from nexusbridge.bybit.api.trade import post_v5_order_cancel_batch
    from nexusbridge.bybit.api.trade import get_v5_order_spot_borrow_check
    from nexusbridge.bybit.api.trade import post_v5_order_disconnected_cancel_all



async def get_fapi_v1_convert_exchange_info(
    self,
    from_asset: str | None = None,
    to_asset: str | None = None,
):
    """
    Query for all convertible token pairs and the tokens’ respective upper/lower limits
    https://developers.binance.com/docs/derivatives/usds-margined-futures/convert
    """
    endpoint = "/fapi/v1/convert/exchangeInfo"
    payload = {}
    if from_asset:
        payload["fromAsset"] = from_asset
    if to_asset:
        payload["toAsset"] = to_asset
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return raw


async def post_fapi_v1_convert_get_quote(
    self, from_asset: str, to_asset: str, **kwargs
):
    """
    Request a quote for the requested token pairs
    https://developers.binance.com/docs/derivatives/usds-margined-futures/convert/Send-quote-request
    """
    endpoint = "/fapi/v1/convert/getQuote"
    payload = {"fromAsset": from_asset, "toAsset": to_asset, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_convert_accept_quote(self, quote_id: str, **kwargs):
    """
    Accept the offered quote by quote ID.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/convert/Accept-Quote
    """
    endpoint = "/fapi/v1/convert/acceptQuote"
    payload = {"quoteId": quote_id, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_convert_order_status(self, **kwargs):
    """
    Query order status by order ID.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/convert/Order-Status
    """
    endpoint = "/fapi/v1/convert/orderStatus"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



import orjson
from typing import Any, Dict, Literal


async def get_v5_asset_delivery_record(
    self, category: Literal["inverse", "linear", "option"], **kwargs
):
    """
    Query delivery records of Invese Futures, USDC Futures and Options, sorted by deliveryTime in descending order
    https://bybit-exchange.github.io/docs/v5/asset/delivery
    """
    endpoint = "/v5/asset/delivery-record"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_settlement_record(self, category: str = "linear", **kwargs):
    """
    Query session settlement records of USDC perpetual and futures
    https://bybit-exchange.github.io/docs/v5/asset/settlement
    """
    endpoint = "/v5/asset/settlement-record"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_exchange_order_record(self, **kwargs):
    """
    Get Coin Exchange Records
    https://bybit-exchange.github.io/docs/v5/asset/exchange
    """
    endpoint = "/v5/asset/exchange/order-record"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_coin_query_info(self, coin: str | None = None):
    """
    Query coin information, including chain information, withdraw and deposit status.
    https://bybit-exchange.github.io/docs/v5/asset/coin-info
    """
    endpoint = "/v5/asset/coin/query-info"
    payload = {}
    if coin:
        payload["coin"] = coin
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_transfer_query_sub_member_list(self):
    """
    Query the sub UIDs under a main UID. It returns up to 2000 sub accounts, if you need more, please call this endpoint.
    https://bybit-exchange.github.io/docs/v5/asset/sub-uid-list
    """
    endpoint = "/v5/asset/transfer/query-sub-member-list"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


# Balances
async def get_v5_asset_transfer_query_asset_info(
    self, account_type: str = "SPOT", coin: str | None = None
):
    """
    Get Asset Info
    https://bybit-exchange.github.io/docs/v5/asset/balance/asset-info
    """
    endpoint = "/v5/asset/transfer/query-asset-info"
    payload = {"account_type": account_type}
    if coin:
        payload["coin"] = coin
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_transfer_query_account_coins_balance(
    self, account_type: str, **kwargs
):
    """
    Get All Coins Balance
    https://bybit-exchange.github.io/docs/v5/asset/balance/all-balance
    """
    endpoint = "/v5/asset/transfer/query-account-coins-balance"
    payload = {"account_type": account_type, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_transfer_query_account_coin_balance(
    self, account_type: str, coin: str, **kwargs
):
    """
    Get Single Coin Balance
    https://bybit-exchange.github.io/docs/v5/asset/balance/account-coin-balance
    """
    endpoint = "/v5/asset/transfer/query-account-coin-balance"
    payload = {"account_type": account_type, "coin": coin, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_withdraw_withdrawable_amount(self, coin: str):
    """
    Get Withdrawable Amount
    https://bybit-exchange.github.io/docs/v5/asset/balance/delay-amount
    """
    endpoint = "/v5/asset/withdraw/withdrawable-amount"
    payload = {"coin": coin}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


# Transfer
async def get_v5_asset_transfer_query_transfer_coin_list(
    self, from_account_type: str, to_account_type: str
):
    """
    Get Transferable Coin
    https://bybit-exchange.github.io/docs/v5/asset/transfer/transferable-coin
    """
    endpoint = "/v5/asset/transfer/query-transfer-coin-list"
    payload = {"fromAccountType": from_account_type, "toAccountType": to_account_type}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_transfer_inter_transfer(
    self,
    from_account_type: str,
    to_account_type: str,
    coin: str,
    transfer_id: str,
    amount: str,
):
    """
    Create Internal Transfer
    https://bybit-exchange.github.io/docs/v5/asset/transfer/create-inter-transfer
    """
    endpoint = "/v5/asset/transfer/inter-transfer"
    payload = {
        "fromAccountType": from_account_type,
        "toAccountType": to_account_type,
        "coin": coin,
        "transferId": transfer_id,
        "amount": amount,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_transfer_query_inter_transfer_list(self, **kwargs):
    """
    Get Internal Transfer Records
    https://bybit-exchange.github.io/docs/v5/asset/transfer/inter-transfer-list
    """
    endpoint = "/v5/asset/transfer/query-inter-transfer-list"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_transfer_universal_transfer(
    self,
    from_account_type: str,
    to_account_type: str,
    coin: str,
    amount: str,
    transfer_id: str,
    from_member_id: int,
    to_member_id: int,
):
    """
    Create Universal Transfer
    https://bybit-exchange.github.io/docs/v5/asset/transfer/unitransfer
    """
    endpoint = "/v5/asset/transfer/universal-transfer"
    payload = {
        "fromAccountType": from_account_type,
        "toAccountType": to_account_type,
        "coin": coin,
        "amount": amount,
        "transferId": transfer_id,
        "fromMemberId": from_member_id,
        "toMemberId": to_member_id,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_transfer_query_universal_transfer_list(self, **kwargs):
    """
    Get Universal Transfer Records
    https://bybit-exchange.github.io/docs/v5/asset/transfer/unitransfer-list
    """
    endpoint = "/v5/asset/transfer/query-universal-transfer-list"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


# Deposit
async def get_v5_asset_deposit_query_allowed_list(self, **kwargs):
    """
    Get Allowed Deposit Coin Info
    https://bybit-exchange.github.io/docs/v5/asset/deposit/deposit-coin-spec
    """
    endpoint = "/v5/asset/deposit/query-allowed-list"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_deposit_to_account(
    self, account_type: Literal["UNIFIED", "SPOT", "CONTRACT", "FUND"]
):
    """
    Set Deposit Account
    https://bybit-exchange.github.io/docs/v5/asset/deposit/set-deposit-acct
    """
    endpoint = "/v5/asset/deposit/deposit-to-account"
    payload = {"accountType": account_type}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_deposit_query_record(self, **kwargs):
    """
    Get Deposit Records (on-chain)
    https://bybit-exchange.github.io/docs/v5/asset/deposit/deposit-record
    """
    endpoint = "/v5/asset/deposit/query-record"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_deposit_query_sub_member_record(
    self, sub_member_id: int, **kwargs
):
    """
    Get Sub Deposit Records (on-chain)
    https://bybit-exchange.github.io/docs/v5/asset/deposit/sub-deposit-record
    """
    endpoint = "/v5/asset/deposit/query-sub-member-record"
    payload = {"subMemberId": sub_member_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_deposit_query_internal_record(self, **kwargs):
    """
    Get Internal Deposit Records (off-chain)
    https://bybit-exchange.github.io/docs/v5/asset/deposit/internal-deposit-record
    """
    endpoint = "/v5/asset/deposit/query-internal-record"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_deposit_query_address(
    self, coin: str, chain_type: str | None = None
):
    """
    Get Master Deposit Address
    https://bybit-exchange.github.io/docs/v5/asset/deposit/master-deposit-addr
    """
    endpoint = "/v5/asset/deposit/query-address"
    payload = {
        "coin": coin,
    }
    if chain_type:
        payload["chainType"] = chain_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_deposit_query_sub_member_address(
    self, coin: str, chain_type: str, sub_member_id: int
):
    """
    Get Sub Deposit Address
    https://bybit-exchange.github.io/docs/v5/asset/deposit/sub-deposit-addr#http-request
    """
    endpoint = "/v5/asset/deposit/query-sub-member-address"
    payload = {"coin": coin, "subMemberId": sub_member_id, "chainType": chain_type}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


# Withdraw
async def get_v5_asset_withdraw_query_record(self, **kwargs):
    """
    Get Withdrawal Records
    https://bybit-exchange.github.io/docs/v5/asset/withdraw/withdraw-record
    """
    endpoint = "/v5/asset/withdraw/query-record"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_withdraw_query_vasp_list(
    self,
):
    """
    This endpoint is particularly used for kyc=KOR users. When withdraw funds, you need to fill entity id.
    https://bybit-exchange.github.io/docs/v5/asset/withdraw/vasp-list
    """
    endpoint = "/v5/asset/withdraw/vasp/list"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_withdraw_create(
    self, address: str, amount: str, timestamp: int, **kwargs
):
    """
    Withdraw assets from your Bybit account. You can make an off-chain transfer if the target wallet address is from Bybit. This means that no blockchain fee will be charged.
    https://bybit-exchange.github.io/docs/v5/asset/withdraw
    """
    endpoint = "/v5/asset/withdraw/create"
    payload = {"address": address, "amount": amount, "timestamp": timestamp, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_withdraw_cancel(
    self,
    _id: str,
):
    """
    Cancel the withdrawal
    https://bybit-exchange.github.io/docs/v5/asset/withdraw/cancel-withdraw
    """
    endpoint = "/v5/asset/withdraw/cancel"
    payload = {"id": _id}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


# Convert
async def get_v5_asset_exchange_query_coin_list(
    self,
    account_type: Literal[
        "eb_convert_funding",
        "eb_convert_uta",
        "eb_convert_spot",
        "eb_convert_contract",
        "eb_convert_inverse",
    ],
    **kwargs,
):
    """
    Get Convert Coin List
    https://bybit-exchange.github.io/docs/v5/asset/convert/convert-coin-list
    """
    endpoint = "/v5/asset/exchange/query-coin-list"
    payload = {"accountType": account_type, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_exchange_quote_apply(
    self,
    from_coin: str,
    to_coin: str,
    request_coin: str,
    request_amount: str,
    account_type: str,
    **kwargs,
):
    """
    Request a Quote
    https://bybit-exchange.github.io/docs/v5/asset/convert/apply-quote
    """
    endpoint = "/v5/asset/exchange/quote-apply"
    payload = {
        "fromCoin": from_coin,
        "toCoin": to_coin,
        "requestCoin": request_coin,
        "requestAmount": request_amount,
        "accountType": account_type,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_v5_asset_exchange_convert_execute(
    self,
    quote_tx_id: str,
):
    """
    Confirm a Quote
    https://bybit-exchange.github.io/docs/v5/asset/convert/confirm-quote
    """
    endpoint = "/v5/asset/exchange/convert-execute"
    payload = {"quoteTxId": quote_tx_id}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_exchange_convert_result_query(
    self, quote_tx_id: str, account_type: str
):
    """
    You can query the exchange result by sending quoteTxId. Make sure you input correct account type and quoteTxId, otherwise you cannot find it.
    https://bybit-exchange.github.io/docs/v5/asset/convert/get-convert-result
    """
    endpoint = "/v5/asset/exchange/convert-result-query"
    payload = {"quoteTxId": quote_tx_id, "accountType": account_type}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_exchange_query_convert_history(self, **kwargs):
    """
    Get Convert history
    https://bybit-exchange.github.io/docs/v5/asset/convert/get-convert-history
    """
    endpoint = "/v5/asset/exchange/query-convert-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



async def post_dapi_v1_listen_key(self, signed=False):
    """
    Start a new user data stream. The stream will close after 60 minutes unless a keepalive is sent. If the account has an active listenKey, that listenKey will be returned and its validity will be extended for 60 minutes.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Start-User-Data-Stream
    """
    endpoint = "/dapi/v1/listenKey"
    raw = await self._fetch("POST", endpoint, signed=signed)
    return raw


async def put_dapi_v1_listen_key(self, signed=False):
    """
    Keepalive a user data stream to prevent a time out. User data streams will close after 60 minutes. It's recommended to send a ping about every 60 minutes.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Keepalive-User-Data-Stream
    """
    endpoint = "/dapi/v1/listenKey"
    raw = await self._fetch("PUT", endpoint, signed=signed)
    return raw


async def delete_dapi_v1_listen_key(self, signed=False):
    """
    Close out a user data stream.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Close-User-Data-Stream
    """
    endpoint = "/dapi/v1/listenKey"
    raw = await self._fetch("DELETE", endpoint, signed=signed)
    return raw



async def post_api_v3_user_data_stream(self):
    endpoint = "/api/v3/userDataStream"

    raw = await self._fetch("POST", endpoint, signed=False)
    return raw


async def put_api_v3_user_data_stream(self, listen_key: str):
    endpoint = "/api/v3/userDataStream"
    payload = {"listenKey": listen_key}

    raw = await self._fetch("PUT", endpoint, payload=payload, signed=False)
    return raw


async def delete_api_v3_user_data_stream(self, listen_key: str):
    endpoint = "/api/v3/userDataStream"
    payload = {"listenKey": listen_key}

    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=False)
    return raw



from typing import Literal, Dict, Any

import orjson


async def get_v5_account_wallet_balance(
    self,
    account_type: Literal["UNIFIED", "CONTRACT", "SPOT"],
    coin: str | None = None,
):
    """
    Get Wallet Balance
    https://bybit-exchange.github.io/docs/v5/account/wallet-balance
    """
    endpoint = "/v5/account/wallet-balance"
    payload = {
        "accountType": account_type,
    }
    if coin:
        payload["coin"] = coin
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_withdrawal(self, coin_name: str):
    """
    Get Transferable Amount (Unified)
    https://bybit-exchange.github.io/docs/v5/account/unified-trans-amnt
    """
    endpoint = "/v5/account/withdrawal"
    payload = {
        "coinName": coin_name,
    }
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_upgrade_to_uta(
    self,
):
    """
    Upgrade to Unified Account
    https://bybit-exchange.github.io/docs/v5/account/upgrade-unified-account
    """
    endpoint = "/v5/account/upgrade-to-uta"
    raw = await self._fetch("POST", endpoint, signed=True)
    return orjson.loads(raw)


async def get_v5_account_borrow_history(self, **kwargs):
    """
    Get Borrow History
    https://bybit-exchange.github.io/docs/v5/account/borrow-history
    """
    endpoint = "/v5/account/borrow-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_quick_repayment(self, coin: str | None = None):
    """
    Repay Liability
    https://bybit-exchange.github.io/docs/v5/account/repay-liability
    """
    endpoint = "/v5/account/quick-repayment"
    payload = {}
    if coin:
        payload["coin"] = coin
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_set_collateral_switch(
    self, coin: str, collateral_switch: Literal["off", "on"]
):
    """
    Set Collateral Coin
    https://bybit-exchange.github.io/docs/v5/account/set-collateral
    """
    endpoint = "/v5/account/set-collateral-switch"
    payload = {
        "coin": coin,
        "collateralSwitch": collateral_switch,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_set_collateral_switch_batch(
    self,
    request: list[Dict[str, Any]],
):
    """
    Batch Set Collateral Coin
    https://bybit-exchange.github.io/docs/v5/account/batch-set-collateral
    """
    endpoint = "/v5/account/set-collateral-switch-batch"
    payload = {
        "request": request,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_collateral_info(self, currency: str | None = None):
    """
    Get Collateral Info
    https://bybit-exchange.github.io/docs/v5/account/collateral-info
    """
    endpoint = "/v5/account/collateral-info"
    payload = {}
    if currency:
        payload["currency"] = currency
    raw = await self._fetch("GET", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def get_v5_asset_coin_greeks(self, base_coin: str | None = None):
    """
    Get Coin Greeks
    https://bybit-exchange.github.io/docs/v5/account/coin-greeks
    """
    endpoint = "/v5/asset/coin-greeks"
    payload = {}
    if base_coin:
        payload["baseCoin"] = base_coin
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_fee_rate(
    self,
    category: Literal["linear", "inverse", "spot", "option"] | None = None,
    **kwargs,
):
    """
    Get Fee Rate
    https://bybit-exchange.github.io/docs/v5/account/fee-rate
    """
    endpoint = "/v5/account/fee-rate"
    payload = {**kwargs}
    if category:
        payload["category"] = category
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_info(self):
    """
    Get Account Info
    https://bybit-exchange.github.io/docs/v5/account/account-info
    """
    endpoint = "/v5/account/info"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def get_v5_account_query_dcp_info(self):
    """
    Get DCP Info
    https://bybit-exchange.github.io/docs/v5/account/dcp-info
    """
    endpoint = "/v5/account/query-dcp-info"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def get_v5_account_transaction_log(self, **kwargs):
    """
    Get Transaction Log
    https://bybit-exchange.github.io/docs/v5/account/transaction-log
    """
    endpoint = "/v5/account/transaction-log"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_contract_transaction_log(self, **kwargs):
    """
    Get Contract Transaction Log
    https://bybit-exchange.github.io/docs/v5/account/contract-transaction-log
    """
    endpoint = "/v5/account/contract-transaction-log"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_smp_group(self):
    """
    Get SMP Group ID
    https://bybit-exchange.github.io/docs/v5/account/smp-group
    """
    endpoint = "/v5/account/smp-group"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def post_v5_account_set_margin_mode(
    self,
    set_margin_mode: Literal["ISOLATED_MARGIN", "REGULAR_MARGIN", "PORTFOLIO_MARGIN"],
):
    """
    Set Margin Mode
    https://bybit-exchange.github.io/docs/v5/account/set-margin-mode
    """
    endpoint = "/v5/account/set-margin-mode"
    payload = {
        "setMarginMode": set_margin_mode,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_set_hedging_mode(
    self, set_hedging_mode: Literal["ON", "OFF"]
):
    """
    Set Spot Hedging
    https://bybit-exchange.github.io/docs/v5/account/set-spot-hedge
    """
    endpoint = "/v5/account/set-hedging-mode"
    payload = {
        "setHedgingMode": set_hedging_mode,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_mmp_modify(
    self,
    base_coin: str,
    window: str,
    frozen_period: str,
    qty_limit: str,
    delta_limit: str,
):
    """
    Set MMP
    https://bybit-exchange.github.io/docs/v5/account/set-mmp
    """
    endpoint = "/v5/account/mmp-modify"
    payload = {
        "baseCoin": base_coin,
        "window": window,
        "frozenPeriod": frozen_period,
        "qtyLimit": qty_limit,
        "deltaLimit": delta_limit,
    }
    raw = await self._fetch("POST", endpoint, payload, signed=True)
    return orjson.loads(raw)


async def post_v5_account_mmp_reset(self, base_coin: str):
    """
    Reset MMP
    https://bybit-exchange.github.io/docs/v5/account/reset-mmp
    """
    endpoint = "/v5/account/mmp-reset"
    payload = {
        "baseCoin": base_coin,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_v5_account_mmp_state(self, base_coin: str):
    """
    Get MMP State
    https://bybit-exchange.github.io/docs/v5/account/get-mmp-state
    """
    endpoint = "/v5/account/mmp-state"
    payload = {
        "baseCoin": base_coin,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



import asyncio

import aiohttp

from typing import Any, Dict, Callable
from urllib.parse import urljoin, urlencode

import orjson
from aiolimiter import AsyncLimiter

from nexusbridge.base import ApiClient, WSClient
from nexusbridge.binance.authentication import rsa_signature, hmac_hashing, ed_25519
from nexusbridge.binance.constants import KeyType
from nexusbridge.binance.error import BinanceClientError, BinanceServerError


class BinanceApiClient(ApiClient):
    def __init__(
        self,
        base_url: str,
        private_key=None,
        private_key_passphrase=None,
        secret: str = None,
        key: str = None,
        timeout: int = 10,
        max_rate: int | None = None,
    ):
        super().__init__(
            secret=secret,
            timeout=timeout,
            max_rate=max_rate,
        )
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

        if key:
            self._headers["X-MBX-APIKEY"] = key

        self.private_key = private_key
        self.private_key_pass = private_key_passphrase
        self.secret = secret
        self.base_url = base_url

    def _get_sign(self, payload):
        if self.private_key:
            return rsa_signature(self.private_key, payload, self.private_key_pass)
        return hmac_hashing(self.secret, payload)

    def _prepare_payload(self, payload: Dict[str, Any], signed: bool) -> str:
        """Prepare payload by encoding and optionally signing."""
        # payload = {k: str(v).lower() if isinstance(v, bool) else v for k, v in payload.items()}
        if signed:
            payload["timestamp"] = self._clock.timestamp_ms()
            encoded_payload = urlencode(payload)
            signature = self._get_sign(encoded_payload)
            return f"{encoded_payload}&signature={signature}"
        return urlencode(payload)

    async def _fetch(
        self,
        method: str,
        endpoint: str,
        payload: Dict[str, Any] = None,
        signed: bool = False,
    ):
        """Make an asynchronous HTTP request."""
        if self._limiter:
            await self._limiter.wait()

        self._init_session()

        url = urljoin(self.base_url, endpoint)
        payload = payload or {}
        encoded_payload = self._prepare_payload(payload, signed)

        if method.upper() == "GET":
            url = f"{url}?{encoded_payload}"
            data = None
        else:
            data = encoded_payload

        self._log.debug(f"Request: {method} {url}")

        try:
            response = await self._session.request(
                method=method,
                url=url,
                headers=self._headers,
                data=data,
            )
            status = response.status
            raw = await response.read()
            self._log.debug(f"Response: {raw}")
            if 400 <= status < 500:
                text = await response.text()
                raise BinanceClientError(status, text, self._headers)
            elif status >= 500:
                text = await response.text()
                raise BinanceServerError(status, text, self._headers)
            return orjson.loads(raw)

        except aiohttp.ClientError as e:
            self._log.error(f"Client Error {method} Url: {url} {e}")
            raise
        except asyncio.TimeoutError:
            self._log.error(f"Timeout {method} Url: {url}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} {e}")
            raise


class BinanceWSClient(WSClient):
    def __init__(
        self,
        url: str,
        handler: Callable[..., Any],
        key=None,
        secret=None,
        listen_key=None,
        private_key=None,
        key_type: KeyType = None,
        private_key_passphrase=None,
        **kwargs,
    ):
        self.key_type = key_type
        self.key = key
        self.secret = secret
        self.private_key = private_key
        self.private_key_passphrase = private_key_passphrase
        self.key_type = (key_type,)
        listen_key = listen_key
        super().__init__(
            url,
            limiter=AsyncLimiter(max_rate=3, time_period=1),
            handler=handler,
            **kwargs,
        )

    def _get_sign(self, payload):
        pairs = payload.split("&")
        pairs.sort()
        payload = "&".join(pairs)
        if self.key_type[0] == KeyType.RSA:
            return rsa_signature(self.private_key, payload, self.private_key_passphrase)
        elif self.key_type[0] == KeyType.Ed:
            return ed_25519(self.private_key, payload, self.private_key_passphrase)
        else:
            return hmac_hashing(self.secret, payload)

    async def _subscribe(
        self, params: str | list[str], subscription_id: str | None = None
    ):
        if isinstance(params, str):
            params = [params]

        new_params = [param for param in params if param not in self._subscriptions]
        if not new_params:
            return

        subscription_id = subscription_id or self._clock.timestamp_ms()
        await self._connect()
        payload = {
            "method": "SUBSCRIBE",
            "params": new_params,
            "id": subscription_id,
        }
        await self._send(payload)
        self._subscriptions.update({param: payload for param in new_params})
        self._log.info(f"Subscribed to {subscription_id} with payload: {payload}")

    async def _unsubscribe(self, params: str | list[str], subscription_id: int):
        if isinstance(params, str):
            params = [params]

        valid_params = [param for param in params if param in self._subscriptions]
        if not valid_params:
            return

        await self._connect()
        payload = {
            "method": "UNSUBSCRIBE",
            "params": valid_params,
            "id": subscription_id,
        }
        await self._send(payload)
        for param in valid_params:
            self._subscriptions.pop(param, None)
        self._log.info(f"Unsubscribed from {subscription_id} with payload: {payload}")

    async def send(
        self, method: str, params: list[Any] | None = None, _id: str | None = None
    ):
        await self._connect()
        payload = {
            "method": method,
            "id": _id or self._clock.timestamp_ms(),
        }
        if params:
            payload["params"] = params

        await self._send(payload)

    async def _resubscribe(self):
        for _, payload in self._subscriptions.items():
            await self._send(payload)



import orjson
from typing import Literal


async def get_v5_market_time(self):
    """
    Get Bybit Server Time
    https://bybit-exchange.github.io/docs/v5/market/time
    """
    endpoint = "/v5/market/time"
    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_v5_market_kline(self, symbol: str, interval: str, **kwargs):
    """
    Query for historical klines (also known as candles/candlesticks). Charts are returned in groups based on the requested interval.
    https://bybit-exchange.github.io/docs/v5/market/kline
    """
    endpoint = "/v5/market/kline"
    payload = {"symbol": symbol, "interval": interval, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_mark_price_kline(self, symbol: str, interval: str, **kwargs):
    """
    Query for historical mark price klines. Charts are returned in groups based on the requested interval.
    https://bybit-exchange.github.io/docs/v5/market/mark-kline
    """
    endpoint = "/v5/market/mark-price-kline"
    payload = {"symbol": symbol, "interval": interval, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_index_price_kline(self, symbol: str, interval: str, **kwargs):
    """
    Query for historical index price klines. Charts are returned in groups based on the requested interval.
    https://bybit-exchange.github.io/docs/v5/market/index-kline
    """
    endpoint = "/v5/market/index-price-kline"
    payload = {"symbol": symbol, "interval": interval, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_premium_index_price_kline(
    self, symbol: str, interval: str, **kwargs
):
    """
    Query for historical premium index klines. Charts are returned in groups based on the requested interval.
    https://bybit-exchange.github.io/docs/v5/market/premium-index-kline
    """
    endpoint = "/v5/market/premium-index-price-kline"
    payload = {"symbol": symbol, "interval": interval, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_instruments_info(self, **kwargs):
    """
    Query for the instrument specification of online trading pairs.
    https://bybit-exchange.github.io/docs/v5/market/instrument
    """
    endpoint = "/v5/market/instruments-info"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_orderbook(
    self,
    symbol: str,
    category: Literal["spot", "linear", "inverse", "option"],
    limit: int | None = None,
):
    """
    Query for orderbook depth data.
    https://bybit-exchange.github.io/docs/v5/market/orderbook
    """
    endpoint = "/v5/market/orderbook"
    payload = {
        "symbol": symbol,
        "category": category,
    }
    if limit:
        payload["limit"] = limit
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_tickers(
    self, category: Literal["spot", "linear", "inverse", "option"], **kwargs
):
    """
    Query for the latest price snapshot, best bid/ask price, and trading volume in the last 24 hours.
    https://bybit-exchange.github.io/docs/v5/market/tickers
    """
    endpoint = "/v5/market/tickers"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_funding_history(
    self, symbol: str, category: Literal["linear", "inverse"], **kwargs
):
    """
    Query for historical funding rates. Each symbol has a different funding interval. For example, if the interval is 8 hours and the current time is UTC 12, then it returns the last funding rate, which settled at UTC 8.
    https://bybit-exchange.github.io/docs/v5/market/history-fund-rate
    """
    endpoint = "/v5/market/funding/history"
    payload = {"symbol": symbol, "category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_recent_trade(
    self, category: Literal["spot", "linear", "inverse", "option"], **kwargs
):
    """
    Query recent public trading data in Bybit.
    https://bybit-exchange.github.io/docs/v5/market/recent-trade
    """
    endpoint = "/v5/market/recent-trade"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_open_interest(
    self,
    symbol: str,
    category: Literal["linear", "inverse"],
    interval_time: Literal["5min", "15min", "30min", "1h", "4h", "1d"],
    **kwargs,
):
    """
    Get the open interest of each symbol.
    https://bybit-exchange.github.io/docs/v5/market/open-interest
    """
    endpoint = "/v5/market/open-interest"
    payload = {
        "symbol": symbol,
        "category": category,
        "interval_time": interval_time,
        **kwargs,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_historical_volatility(self, category: str = "option", **kwargs):
    """
    Query option historical volatility
    https://bybit-exchange.github.io/docs/v5/market/iv
    """
    endpoint = "/v5/market/historical-volatility"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_insurance(
    self,
    coin: str | None = None,
):
    """
    Query for Bybit insurance pool data (BTC/USDT/USDC etc). The data is updated every 24 hours.
    https://bybit-exchange.github.io/docs/v5/market/insurance
    """
    endpoint = "/v5/market/insurance"
    payload = {}
    if coin:
        payload["coin"] = coin
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_risk_limit(
    self, category: Literal["linear", "inverse"], **kwargs
):
    """
    Query for the risk limit.
    https://bybit-exchange.github.io/docs/v5/market/risk-limit
    """
    endpoint = "/v5/market/risk-limit"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_delivery_price(
    self, category: Literal["linear", "inverse", "option"], **kwargs
):
    """
    Get the delivery price.
    https://bybit-exchange.github.io/docs/v5/market/delivery-price
    """
    endpoint = "/v5/market/delivery-price"
    payload = {"category": category, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_v5_market_account_ratio(
    self,
    category: Literal["linear", "inverse"],
    symbol: str,
    period: Literal["5min", "15min", "30min", "1h", "4h", "1d"],
    **kwargs,
):
    """
     Get Long Short Ratio
    https://bybit-exchange.github.io/docs/v5/market/long-short-ratio
    """
    endpoint = "/v5/market/account-ratio"
    payload = {"category": category, "symbol": symbol, "period": period, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)



from typing import Literal

from nexusbridge.binance.base import BinanceApiClient
from nexusbridge.binance.constants import BinanceUrl, ALL_URL, BinanceInstrumentType


class MarginTradingApi(BinanceApiClient):
    def __init__(
        self,
        key=None,
        secret=None,
        binance_url: Literal[BinanceUrl.MAIN, BinanceUrl.TEST] = BinanceUrl.MAIN,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            kwargs["base_url"] = ALL_URL[BinanceInstrumentType.MARGIN].get_url(
                binance_url
            )
        super().__init__(key=key, secret=secret, **kwargs)

    from nexusbridge.binance.margin.account import get_sapi_v1_simpleEarn_flexible_list
    from nexusbridge.binance.margin.account import get_sapi_v1_simpleEarn_locked_list
    from nexusbridge.binance.margin.history import (
        get_sapi_v1_simpleEarn_flexible_history_rateHistory,
    )
    from nexusbridge.binance.margin.history import (
        get_sapi_v1_simpleEarn_flexible_history_rewardsRecord,
    )
    from nexusbridge.binance.margin.history import (
        get_sapi_v1_simpleEarn_locked_history_rewardsRecord,
    )



from typing import Literal


async def get_sapi_v1_simpleEarn_flexible_history_rateHistory(
    self,
    productId: str,
    aprPeriod: str | None = None,
    startTime: int | None = None,
    endTime: int | None = None,
    current: int | None = None,
    size: int | None = None,
):
    """
    https://developers.binance.com/docs/simple_earn/history/Get-Rate-History
    """
    endpoint = "/sapi/v1/simple-earn/flexible/history/rateHistory"
    payload = {
        "productId": productId,
        "aprPeriod": aprPeriod,
        "startTime": startTime,
        "endTime": endTime,
        "current": current,
        "size": size,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_sapi_v1_simpleEarn_flexible_history_rewardsRecord(
    self,
    type: Literal["Bonus", "REALTIME", "ALL"],
    productId: str | None = None,
    asset: str | None = None,
    startTime: int | None = None,
    endTime: int | None = None,
):
    """
    /sapi/v1/simple-earn/flexible/history/rewardsRecord
    """
    endpoint = "/sapi/v1/simple-earn/flexible/history/rewardsRecord"
    payload = {
        "type": type,
        "productId": productId,
        "asset": asset,
        "startTime": startTime,
        "endTime": endTime,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_sapi_v1_simpleEarn_locked_history_rewardsRecord(
    self,
    positionId: int | None = None,
    asset: str | None = None,
    startTime: int | None = None,
    endTime: int | None = None,
    current: int | None = None,
    size: int | None = None,
):
    """
    /sapi/v1/simple-earn/locked/history/rewardsRecord

    Request Weight(IP)
    150

    Request Parameters
    Name	Type	Mandatory	Description
    positionId	INT	NO
    asset	STRING	NO
    startTime	LONG	NO
    endTime	LONG	NO
    current	LONG	NO	Currently querying the page. Start from 1, Default:1, Max: 1,000
    size	LONG	NO	Default:10, Max:100
    recvWindow	LONG	NO
    timestamp	LONG	YES

    """
    endpoint = "/sapi/v1/simple-earn/locked/history/rewardsRecord"
    payload = {
        "positionId": positionId,
        "asset": asset,
        "startTime": startTime,
        "endTime": endTime,
        "current": current,
        "size": size,
    }
    payload = {k: v for k, v in payload.items() if v is not None}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



import json
from typing import Literal

from nexusbridge.binance.constants import OrderType, Side


async def post_fapi_v1_order(self, symbol: str, side: Side, _type: OrderType, **kwargs):
    """
    Send in a new order.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api
    """
    endpoint = "/fapi/v1/order"
    payload = {"symbol": symbol, "side": side.value, "type": _type.value, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_batch_order(self, batch_orders: list[dict], **kwargs):
    """
    Place Multiple Orders
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Place-Multiple-Orders
    """
    endpoint = "/fapi/v1/batchOrders"
    payload = {"batchOrders": json.dumps(batch_orders), **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def put_fapi_v1_order(
    self, symbol: str, side: Side, quantity: float, price: float, **kwargs
):
    """
    Order modify function, currently only LIMIT order modification is supported, modified orders will be reordered in the match queue
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Modify-Order
    """
    endpoint = "/fapi/v1/order"
    payload = {
        "symbol": symbol,
        "side": side.value,
        "quantity": quantity,
        "price": price,
        **kwargs,
    }
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def put_fapi_v1_batch_order(self, batch_orders: list[dict], **kwargs):
    """
    Modify Multiple Orders (TRADE)
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Modify-Multiple-Orders
    """
    endpoint = "/fapi/v1/batchOrders"
    payload = {"batchOrders": json.dumps(batch_orders), **kwargs}
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_order_amendment(self, symbol: str, **kwargs):
    """
    Get order modification history
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Get-Order-Modify-History
    """
    endpoint = "/fapi/v1/orderAmendment"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def delete_fapi_v1_order(self, symbol: str, **kwargs):
    """
    Cancel an active order.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Cancel-Order
    """
    endpoint = "/fapi/v1/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_fapi_v1_batch_order(
    self,
    symbol: str,
    order_id_list: list[int] | None = None,
    orig_client_order_id_list: list[int] | None = None,
    **kwargs,
):
    """
    Cancel Multiple Orders
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Cancel-Multiple-Orders#request-parameters
    """
    endpoint = "/fapi/v1/batchOrders"
    payload = {"symbol": symbol, **kwargs}
    if order_id_list:
        payload["orderIdList"] = json.dumps(order_id_list)
    if orig_client_order_id_list:
        payload["origClientOrderIdList"] = json.dumps(orig_client_order_id_list)
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_fapi_v1_all_open_orders(self, symbol: str, **kwargs):
    """
    Cancel All Open Orders
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Cancel-All-Open-Orders
    """
    endpoint = "/fapi/v1/allOpenOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_countdown_cancel_all(
    self, symbol: str, countdown_time: int, **kwargs
):
    """
    Cancel all open orders of the specified symbol at the end of the specified countdown. The endpoint should be called repeatedly as heartbeats so that the existing countdown time can be canceled and replaced by a new one.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Auto-Cancel-All-Open-Orders
    """
    endpoint = "/fapi/v1/countdownCancelAll"
    payload = {"symbol": symbol, "countdownTime": countdown_time, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_order(self, symbol: str, **kwargs):
    """
    Check an order's status.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Query-Order
    """
    endpoint = "/fapi/v1/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_all_orders(self, symbol: str, **kwargs):
    """
    Get all account orders; active, canceled, or filled.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/All-Orders
    """
    endpoint = "/fapi/v1/allOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_open_orders(self, **kwargs):
    """
    Get all open orders on a symbol.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Current-All-Open-Orders
    """
    endpoint = "/fapi/v1/openOrders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_open_order(self, symbol: str, **kwargs):
    """
    Query open order
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Query-Current-Open-Order
    """
    endpoint = "/fapi/v1/openOrder"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_force_orders(self, **kwargs):
    """
    Query user's Force Orders
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Users-Force-Orders
    """
    endpoint = "/fapi/v1/forceOrders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_user_trades(self, symbol: str, **kwargs):
    """
    Get trades for a specific account and symbol.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Account-Trade-List
    """
    endpoint = "/fapi/v1/userTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_margin_type(
    self, symbol: str, margin_type: Literal["ISOLATED", "CROSSED"], **kwargs
):
    """
    Change symbol level margin type
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Change-Margin-Type
    """
    endpoint = "/fapi/v1/marginType"
    payload = {"symbol": symbol, "marginType": margin_type, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_position_side_dual(
    self, dual_side_position: Literal["BOTH", "LONG", "SHORT"], **kwargs
):
    """
    Change user's position mode (Hedge Mode or One-way Mode ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Change-Position-Mode
    """
    endpoint = "/fapi/v1/positionSide/dual"
    payload = {"dualSidePosition": dual_side_position, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_leverage(self, symbol: str, leverage: int, **kwargs):
    """
    Change user's initial leverage of specific symbol market.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Change-Initial-Leverage
    """
    endpoint = "/fapi/v1/leverage"
    payload = {"symbol": symbol, "leverage": leverage, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_multi_assets_margin(self, multi_assets_margin: str, **kwargs):
    """
    Change user's Multi-Assets mode (Multi-Assets Mode or Single-Asset Mode) on Every symbol
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Change-Multi-Assets-Mode
    """
    endpoint = "/fapi/v1/multiAssetsMargin"
    payload = {"multiAssetsMargin": multi_assets_margin, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_position_margin(
    self, symbol: str, amount: float, _type: Literal[1, 2], **kwargs
):
    """
    Modify Isolated Position Margin
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Modify-Isolated-Position-Margin
    """
    endpoint = "/fapi/v1/positionMargin"
    payload = {"symbol": symbol, "amount": amount, "type": _type, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v2_position_risk(self, **kwargs):
    """
    Get current position information.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Position-Information-V2
    """
    endpoint = "/fapi/v2/positionRisk"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v3_position_risk(self, **kwargs):
    """
    Get current position information(only symbol that has position or open orders will be returned).
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Position-Information-V3
    """
    endpoint = "/fapi/v3/positionRisk"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_adl_quantile(self, **kwargs):
    """
    Position ADL Quantile Estimation
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Position-ADL-Quantile-Estimation
    """
    endpoint = "/fapi/v1/adlQuantile"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_fapi_v1_position_margin_history(self, symbol: str, **kwargs):
    """
    Get position margin change history
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Get-Position-Margin-Change-History
    """
    endpoint = "/fapi/v1/positionMargin/history"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_fapi_v1_order_test(
    self, symbol: str, side: Side, _type: OrderType, **kwargs
):
    """
    Testing order request, this order will not be submitted to matching engine
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/New-Order-Test
    """
    endpoint = "/fapi/v1/order/test"
    payload = {"symbol": symbol, "side": side.value, "type": _type.value, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw



from typing import Literal

import orjson


async def post_api_v5_trading_bot_signal_create_signal(
    self,
    signal_chan_name: str,
    signal_chan_desc: str | None = None,
):
    """
    Create a signal.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-signal-bot-trading
    """
    endpoint = "/api/v5/tradingBot/signal/create-signal"
    payload = {
        "signalChanName": signal_chan_name,
    }
    if signal_chan_desc:
        payload["signalChanDesc"] = signal_chan_desc
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


# POST /api/v5/tradingBot/recurring/order-algo
async def post_api_v5_trading_bot_recurring_order_algo(
    self,
    stgy_name: str,
    recurring_list: list[dict[str, str]],
    period: Literal["monthly", "weekly", "daily", "hourly"],
    recurring_time: str,
    time_zone: str,
    amt: str,
    investment_ccy: str,
    td_mode: Literal["cross", "cash"],
    **kwargs,
):
    """
    Create a recurring order algo.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-post-place-recurring-buy-order
    """
    endpoint = "/api/v5/tradingBot/recurring/order-algo"
    payload = {
        "stgyName": stgy_name,
        "recurringList": recurring_list,
        "period": period,
        "recurringTime": recurring_time,
        "timeZone": time_zone,
        "amt": amt,
        "investmentCcy": investment_ccy,
        "tdMode": td_mode,
        **kwargs,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_recurring_amend_order_algo(
    self,
    algo_id: str,
    stgy_name: str,
):
    """
    Amend recurring buy order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-post-amend-recurring-buy-order
    """
    endpoint = "/api/v5/tradingBot/recurring/amend-order-algo"
    payload = {
        "algoId": algo_id,
        "stgyName": stgy_name,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trading_bot_recurring_stop_order_algo(
    self,
    algo_id: list[str],
):
    """
    Stop recurring buy order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-post-stop-recurring-buy-order
    """
    endpoint = "/api/v5/tradingBot/recurring/stop-order-algo"
    payload = []
    for algo in algo_id:
        payload.append({"algoId": algo})

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_recurring_orders_algo_pending(self, **kwargs):
    """
    Get pending recurring orders
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-get-recurring-buy-order-list
    """
    endpoint = "/api/v5/tradingBot/recurring/orders-algo-pending"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_recurring_orders_algo_history(self, **kwargs):
    """
    Get historical recurring orders
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-get-recurring-buy-order-history
    """
    endpoint = "/api/v5/tradingBot/recurring/orders-algo-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_recurring_orders_algo_details(self, algo_id: str):
    """
    Get recurring order details
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-get-recurring-buy-order-details
    """
    endpoint = "/api/v5/tradingBot/recurring/orders-algo-details"
    payload = {"algoId": algo_id}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trading_bot_recurring_sub_orders(self, algo_id: str, **kwargs):
    """
    Get recurring sub orders
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-recurring-buy-get-recurring-buy-sub-orders
    """
    endpoint = "/api/v5/tradingBot/recurring/sub-orders"
    payload = {"algoId": algo_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



from enum import Enum


class BaseUrl:
    def __init__(
        self,
        testnet_rest_url: str | None,
        main_rest_url: str,
        main_ws_url: str | None,
        testnet_ws_url: str | None,
        main_ws_stream_url: str | None = None,
    ):
        self.testnet_rest_url = testnet_rest_url
        self.main_rest_url = main_rest_url
        self.main_ws_url = main_ws_url
        self.testnet_ws_url = testnet_ws_url
        self.main_ws_stream_url = main_ws_stream_url

    def get_url(self, url_type: "BinanceUrl") -> str:
        """根据 BinanceUrl 枚举获取对应的 URL"""
        if self.main_ws_stream_url is None and url_type == BinanceUrl.WS_STREAM:
            raise ValueError("main_ws_stream_url is not set")

        url_mapping = {
            BinanceUrl.MAIN: self.main_rest_url,
            BinanceUrl.TEST: self.testnet_rest_url,
            BinanceUrl.WS: self.main_ws_url,
            BinanceUrl.TEST_WS: self.testnet_ws_url,
            BinanceUrl.WS_STREAM: self.main_ws_stream_url,
        }
        return url_mapping[url_type]


class BinanceUrl(Enum):
    MAIN = 0
    TEST = 1
    WS = 2
    TEST_WS = 3
    WS_STREAM = 4


class KeyType(Enum):
    HMAC = 0  # "HMAC-SHA-256"
    RSA = 1  # "RSA"
    Ed = 2  # "Ed25519"


class BinanceInstrumentType(Enum):
    Derivatives_PM = "Derivatives_PM"
    Derivatives_UM = "Derivatives_UM"
    Derivatives_CM = "Derivatives_CM"
    SPOT = "SPOT"
    MARGIN = "MARGIN"


ALL_URL = {
    BinanceInstrumentType.Derivatives_UM: BaseUrl(
        testnet_rest_url="https://testnet.binancefuture.com",
        testnet_ws_url="wss://testnet.binancefuture.com/ws-fapi/v1",
        main_rest_url="https://fapi.binance.com",
        main_ws_url="wss://ws-fapi.binance.com/ws-fapi/v1",
        main_ws_stream_url="wss://fstream.binance.com/ws",
    ),
    BinanceInstrumentType.Derivatives_CM: BaseUrl(
        testnet_rest_url="https://testnet.binancefuture.com/ws",
        testnet_ws_url="wss://dstream.binancefuture.com/ws",
        main_rest_url="https://dapi.binance.com",
        main_ws_url="wss://ws-dapi.binance.com/ws-dapi/v1",
        main_ws_stream_url="wss://dstream.binance.com/ws",
    ),
    BinanceInstrumentType.SPOT: BaseUrl(
        testnet_rest_url="https://testnet.binance.vision/api",
        testnet_ws_url="wss://testnet.binance.vision/ws-api/v3",
        main_rest_url="https://api.binance.com",
        main_ws_url="wss://ws-api.binance.com:9443/ws-api/v3",
        main_ws_stream_url="wss://stream.binance.com:9443/ws",
    ),
    BinanceInstrumentType.Derivatives_PM: BaseUrl(
        testnet_rest_url=None,
        testnet_ws_url=None,
        main_rest_url="https://papi.binance.com",
        main_ws_url="wss://fstream.binance.com/pm",
    ),
    BinanceInstrumentType.MARGIN: BaseUrl(
        testnet_rest_url=None,
        testnet_ws_url=None,
        main_rest_url="https://api.binance.com",
        main_ws_url="wss://ws-api.binance.com:9443/ws-api/v3",
        main_ws_stream_url="wss://stream.binance.com:9443/ws",
    ),
}


class Interval(Enum):
    MINUTE_1 = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"
    HOUR_1 = "1h"
    HOUR_2 = "2h"
    HOUR_4 = "4h"
    HOUR_6 = "6h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"
    DAY_1 = "1d"
    DAY_3 = "3d"
    WEEK_1 = "1w"
    MONTH_1 = "1M"


class Side(Enum):
    BUY = "BUY"
    SELL = "SELL"


class OrderType(Enum):
    LIMIT = "LIMIT"
    MARKET = "MARKET"
    STOP_LOSS = "STOP_LOSS"
    STOP_LOSS_LIMIT = "STOP_LOSS_LIMIT"
    TAKE_PROFIT = "TAKE_PROFIT"
    TAKE_PROFIT_LIMIT = "TAKE_PROFIT_LIMIT"
    LIMIT_MAKER = "LIMIT_MAKER"


class ContractType(Enum):
    PERPETUAL = "PERPETUAL"
    CURRENT_QUARTER = "CURRENT_QUARTER"
    NEXT_QUARTER = "NEXT_QUARTER"
    ALL = "ALL"



from typing import Literal
from urllib.parse import urlencode

from nexusbridge.binance.authentication import hmac_hashing, rsa_signature, ed_25519
from nexusbridge.binance.constants import Interval, OrderType


async def agg_trade(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    The Aggregate Trade Streams push trade information that is aggregated for a single taker order.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#aggregate-trade-streams
    """
    stream_name = "{}@aggTrade".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def trade(self, symbol: str, subscription: bool = True, _id: int | None = None):
    """
    The Trade Streams push raw trade information; each trade has a unique buyer and seller.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#trade-streams
    """
    stream_name = "{}@trade".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def kline(
    self,
    symbol: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    The Kline/Candlestick Stream push updates to the current klines/candlestick every second in UTC+0 timezone
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#klinecandlestick-streams-for-utc
    """
    stream_name = "{}@kline_{}".format(symbol.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ui_kline(
    self,
    symbol: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    The Kline/Candlestick Stream push updates to the current klines/candlestick every second in UTC+8 timezone
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#klinecandlestick-streams-with-timezone-offset
    """
    stream_name = "{}@kline_{}@+08:00".format(symbol.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    24hr rolling window mini-ticker statistics. These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#individual-symbol-mini-ticker-stream
    """
    stream_name = "{}@miniTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window mini-ticker statistics for all symbols that changed in an array. These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#all-market-mini-tickers-stream
    """
    stream_name = "!miniTicker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker(self, symbol: str, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for a single symbol. These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#individual-symbol-ticker-streams
    """
    stream_name = "{}@ticker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for all symbols that changed in an array. These are NOT the statistics of the UTC day, but a 24hr rolling window for the previous 24hrs. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#all-market-tickers-stream
    """
    stream_name = "!ticker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker_window(
    self,
    symbol: str,
    window_size: Literal["1h", "4h", "1d"],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Rolling window ticker statistics for a single symbol, computed over multiple windows.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#individual-symbol-rolling-window-statistics-streams
    """
    stream_name = "{}@ticker_{}".format(symbol.lower(), window_size)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker_window_arr(
    self,
    window_size: Literal["1h", "4h", "1d"],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Rolling window ticker statistics for all market symbols, computed over multiple windows. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#all-market-rolling-window-statistics-streams
    """
    stream_name = "!ticker_{}@arr".format(window_size)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def book_ticker(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Pushes any update to the best bid or ask's price or quantity in real-time for a specified symbol. Multiple <symbol>@bookTicker streams can be subscribed to over one connection.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#individual-symbol-book-ticker-streams
    """
    stream_name = "{}@bookTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def avg_price(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Average price streams push changes in the average price over a fixed time interval.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#average-price
    """
    stream_name = "{}@avgPrice".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def depth(
    self,
    symbol: str,
    levels: Literal[5, 10, 20],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Top <levels> bids and asks, pushed every second. Valid <levels> are 5, 10, or 20.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#partial-book-depth-streams
    """
    stream_name = "{}@depth{}".format(symbol.lower(), levels)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def diff_depth(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Order book price and quantity depth updates used to locally manage an order book.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-streams#diff-depth-stream
    """
    stream_name = "{}@depth".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def order_book_api(self, symbol: str, limit: int = 100, _id: int | None = None):
    """
    Get current order book.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests
    """
    payload = {"symbol": symbol, "limit": limit}
    await self.send("depth", params=payload, _id=_id)
    return


async def recent_trades_api(
    self, symbol: str, limit: int = 100, _id: int | None = None
):
    """
    Get recent trades.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#recent-trades
    """
    payload = {"symbol": symbol, "limit": limit}
    await self.send("trades.recent", params=payload, _id=_id)
    return


async def historical_trades_api(self, symbol: str, _id: int | None = None, **kwargs):
    """
    Get historical trades.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#historical-trades
    """
    payload = {"symbol": symbol, **kwargs}

    return await self.send("trades.historical", params=payload, _id=_id)


async def agg_trades_api(self, symbol: str, _id: int | None = None, **kwargs):
    """
    Get aggregate trades.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#aggregate-trades
    """
    payload = {"symbol": symbol, **kwargs}
    return await self.send("trades.aggregate", params=payload, _id=_id)


async def klines_api(
    self, symbol: str, interval: Interval, _id: int | None = None, **kwargs
):
    """
    Get klines (candlestick bars).
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#klines
    """
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    return await self.send("klines", params=payload, _id=_id)


async def ui_klines_api(
    self, symbol: str, interval: Interval, _id: int | None = None, **kwargs
):
    """
    Get klines (candlestick bars) optimized for presentation.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#ui-klines
    """
    payload = {"symbol": symbol, "interval": interval.value, **kwargs}
    return await self.send("uiKlines", params=payload, _id=_id)


async def avg_price_api(
    self,
    symbol: str,
    _id: int | None = None,
):
    """
    Get current average price for a symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#current-average-price
    """
    payload = {
        "symbol": symbol,
    }
    return await self.send("avgPrice", params=payload, _id=_id)


async def ticker_24hr_api(
    self,
    symbol: str,
    _id: int | None = None,
):
    """
    Get 24-hour rolling window price change statistics.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#24hr-ticker-price-change-statistics
    """
    payload = {
        "symbol": symbol,
    }
    return await self.send("ticker.24hr", params=payload, _id=_id)


async def ticker_trading_day_api(
    self,
    symbol: str | list[str],
    _id: int | None = None,
):
    """
    Price change statistics for a trading day.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#trading-day-ticker
    """
    payload = {}
    if isinstance(symbol, str):
        payload["symbol"] = symbol
    else:
        payload["symbols"] = symbol
    return await self.send("ticker.tradingDay", params=payload, _id=_id)


async def ticker_api(self, symbol: str | list[str], _id: int | None = None, **kwargs):
    """
    Get rolling window price change statistics with a custom window.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#rolling-window-price-change-statistics
    """
    payload = {**kwargs}
    if isinstance(symbol, str):
        payload["symbol"] = symbol
    else:
        payload["symbols"] = symbol
    return await self.send("ticker", params=payload, _id=_id)


async def ticker_price_api(
    self,
    symbol: str | list[str],
    _id: int | None = None,
):
    """
    Get the latest market price for a symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#symbol-price-ticker
    """
    payload = {}
    if isinstance(symbol, str):
        payload["symbol"] = symbol
    else:
        payload["symbols"] = symbol
    return await self.send("ticker.price", params=payload, _id=_id)


async def ticker_book_api(
    self,
    symbol: str | list[str],
    _id: int | None = None,
):
    """
    Get the current best price and quantity on the order book.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/market-data-requests#symbol-order-book-ticker
    """
    payload = {}
    if isinstance(symbol, str):
        payload["symbol"] = symbol
    else:
        payload["symbols"] = symbol
    return await self.send("ticker.book", params=payload, _id=_id)


async def session_logon_api(self, _id: int | None = None, **kwargs):
    """
    Authenticate WebSocket connection using the provided API key.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/authentication-requests#log-in-with-api-key-signed
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = ed_25519(
        self.private_key, encoded_payload, self.private_key_passphrase
    )
    return await self.send("session.logon", params=payload, _id=_id)


async def session_status_api(
    self,
    _id: int | None = None,
):
    """
    Query the status of the WebSocket connection, inspecting which API key (if any) is used to authorize requests.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/authentication-requests#query-session-status
    """
    return await self.send("session.status", _id=_id)


async def session_logout_api(
    self,
    _id: int | None = None,
):
    """
    Forget the API key previously authenticated. If the connection is not authenticated, this request does nothing.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/authentication-requests#log-out-of-the-session
    """
    return await self.send("session.logout", _id=_id)


async def order_place_api(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Send in a new order.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests
    """
    payload = {
        "symbol": symbol,
        "side": side,
        "type": _type.value,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.place", params=payload, _id=_id)


async def order_test_api(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Test order placement.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#test-new-order-trade
    """
    payload = {
        "symbol": symbol,
        "side": side,
        "type": _type.value,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.test", params=payload, _id=_id)


async def order_status_api(
    self,
    symbol: str,
    order_id: int | None = None,
    orig_client_order_id: str | None = None,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Check execution status of an order.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#query-order-user_data
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if order_id:
        payload["orderId"] = order_id
    if orig_client_order_id:
        payload["origClientOrderId"] = orig_client_order_id
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.status", params=payload, _id=_id)


async def order_cancel_api(
    self,
    symbol: str,
    order_id: int | None = None,
    orig_client_order_id: str | None = None,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    order.cancel
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#cancel-order-trade
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if order_id:
        payload["orderId"] = order_id
    if orig_client_order_id:
        payload["origClientOrderId"] = orig_client_order_id
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.cancel", params=payload, _id=_id)


async def order_cancel_replace_api(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    cancel_replace_mode: Literal["STOP_ON_FAILURE", "ALLOW_FAILURE"],
    order_id: int | None = None,
    orig_client_order_id: str | None = None,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Cancel an existing order and immediately place a new order instead of the canceled one.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#cancel-and-replace-order-trade
    """
    payload = {
        "symbol": symbol,
        "timestamp": self._clock.timestamp_ms(),
        "side": side,
        "type": _type.value,
        "cancelReplaceMode": cancel_replace_mode,
        **kwargs,
    }
    if order_id:
        payload["orderId"] = order_id
    if orig_client_order_id:
        payload["origClientOrderId"] = orig_client_order_id
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.cancelReplace", params=payload, _id=_id)


async def open_orders_status_api(
    self, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query execution status of all open orders.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#current-open-orders-user_data
    """
    payload = {"timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("openOrders.status", params=payload, _id=_id)


async def open_orders_cancel_all_api(
    self, symbol: str, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Cancel all open orders on a symbol. This includes orders that are part of an order list.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#cancel-open-orders-trade
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("openOrders.cancelAll", params=payload, _id=_id)


async def order_list_place_api(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    price: float,
    _type: OrderType,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Send in a new one-cancels-the-other (OCO) pair: LIMIT_MAKER + STOP_LOSS/STOP_LOSS_LIMIT orders (called legs), where activation of one order immediately cancels the other.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#order-lists
    """
    payload = {
        "symbol": symbol,
        "price": price,
        "side": side,
        "type": _type.value,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("orderList.place", params=payload, _id=_id)


async def order_place(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    quantity: float,
    _id: int | None = None,
    sign: bool = False,
    **kwargs,
):
    """
    Place new order using SOR (TRADE)
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/trading-requests#sor
    """
    payload = {
        "symbol": symbol,
        "side": side,
        "type": _type.value,
        "timestamp": self._clock.timestamp_ms(),
        "quantity": quantity,
        **kwargs,
    }
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.place", params=payload, _id=_id)


async def account_status_api(
    self, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query information about your account.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-information-user_data
    """
    payload = {"timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.status", params=payload, _id=_id)


async def rate_limit_order_api(
    self, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query your current unfilled order count for all intervals.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#unfilled-order-count-user_data
    """
    payload = {"timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.rateLimits.orders", params=payload, _id=_id)


async def all_orders_api(
    self, symbol: str, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query information about all your orders – active, canceled, filled – filtered by time range.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-order-history-user_data
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("allOrders", params=payload, _id=_id)


async def all_order_lists_api(
    self, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query information about all your order lists, filtered by time range.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-order-list-history-user_data
    """
    payload = {"timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("allOrderLists", params=payload, _id=_id)


async def my_trades_api(
    self, symbol: str, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Query information about all your trades, filtered by time range.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-trade-history-user_data
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("myTrades", params=payload, _id=_id)


async def my_prevented_matches_api(
    self, symbol: str, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Displays the list of orders that were expired due to STP.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-prevented-matches-user_data
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("myPreventedMatches", params=payload, _id=_id)


async def my_allocations_api(
    self, symbol: str, _id: int | None = None, sign: bool = False, **kwargs
):
    """
    Retrieves allocations resulting from SOR order placement.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-allocations-user_data
    """
    payload = {"symbol": symbol, "timestamp": self._clock.timestamp_ms(), **kwargs}
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("myAllocations", params=payload, _id=_id)


async def account_commission_api(
    self,
    symbol: str,
    _id: int | None = None,
    sign: bool = False,
):
    """
    Get current account commission rates.
    https://developers.binance.com/docs/binance-spot-api-docs/web-socket-api/account-requests#account-commission-rates-user_data
    """
    payload = {
        "symbol": symbol,
        "timestamp": self._clock.timestamp_ms(),
    }
    if sign:
        payload["apiKey"] = self.key
        encoded_payload = urlencode(payload)
        payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.commission", params=payload, _id=_id)



import json
from typing import Literal

from nexusbridge.binance.constants import Side, OrderType


async def post_dapi_v1_order(self, symbol: str, side: Side, _type: OrderType, **kwargs):
    """

    Send in a new order.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade
    """
    endpoint = "/dapi/v1/order"
    payload = {"symbol": symbol, "side": side.value, "type": _type.value, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_batch_order(self, batch_orders: list[dict], **kwargs):
    """
    Place Multiple Orders
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Place-Multiple-Orders
    """
    endpoint = "/dapi/v1/batchOrders"
    payload = {"batchOrders": json.dumps(batch_orders), **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def put_dapi_v1_order(self, symbol: str, side: Side, **kwargs):
    """
    Order modify function, currently only LIMIT order modification is supported, modified orders will be reordered in the match queue
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Modify-Order
    """
    endpoint = "/dapi/v1/order"
    payload = {"symbol": symbol, "side": side.value, **kwargs}
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def put_dapi_v1_batch_order(self, batch_orders: list[dict], **kwargs):
    """
    Modify Multiple Orders
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Modify-Multiple-Orders
    """
    endpoint = "/dapi/v1/batchOrders"
    payload = {"batchOrders": json.dumps(batch_orders), **kwargs}
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_order_amendment(self, symbol: str, **kwargs):
    """
    Get order modification history
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Get-Order-Modify-History
    """
    endpoint = "/dapi/v1/orderAmendment"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def delete_dapi_v1_order(self, symbol: str, **kwargs):
    """
    Cancel an active order.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/rest-api/Cancel-Order
    """
    endpoint = "/dapi/v1/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_dapi_v1_batch_order(
    self,
    symbol: str,
    order_id_list: list[int] | None = None,
    orig_client_order_id_list: list[int] | None = None,
    **kwargs,
):
    """
    Cancel Multiple Orders
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Cancel-Multiple-Orders
    """
    endpoint = "/dapi/v1/batchOrders"
    payload = {"symbol": symbol, **kwargs}
    if order_id_list:
        payload["orderIdList"] = json.dumps(order_id_list)
    if orig_client_order_id_list:
        payload["origClientOrderIdList"] = json.dumps(orig_client_order_id_list)
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_dapi_v1_all_open_orders(self, symbol: str, **kwargs):
    """
    Cancel All Open Orders
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Cancel-All-Open-Orders
    """
    endpoint = "/dapi/v1/allOpenOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_countdown_cancel_all(
    self, symbol: str, countdown_time: int, **kwargs
):
    """
    Cancel all open orders of the specified symbol at the end of the specified countdown. This rest endpoint means to ensure your open orders are canceled in case of an outage. The endpoint should be called repeatedly as heartbeats so that the existing countdown time can be canceled and repalced by a new one. The system will check all countdowns approximately every 10 milliseconds, so please note that sufficient redundancy should be considered when using this function. We do not recommend setting the countdown time to be too precise or too small.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Auto-Cancel-All-Open-Orders
    """
    endpoint = "/dapi/v1/countdownCancelAll"
    payload = {"symbol": symbol, "countdownTime": countdown_time, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_order(self, symbol: str, **kwargs):
    """
    Check an order's status.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Query-Order
    """
    endpoint = "/dapi/v1/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_all_orders(self, symbol: str, **kwargs):
    """
    Get all account orders; active, canceled, or filled.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/All-Orders
    """
    endpoint = "/dapi/v1/allOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_open_orders(self, **kwargs):
    """
    Get all open orders on a symbol.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Current-All-Open-Orders
    """
    endpoint = "/dapi/v1/openOrders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_open_order(self, symbol: str, **kwargs):
    """
    Query open order
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Query-Current-Open-Order
    """
    endpoint = "/dapi/v1/openOrder"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_force_orders(self, **kwargs):
    """
    Query user's Force Orders
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Users-Force-Orders
    """
    endpoint = "/dapi/v1/forceOrders"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_user_trades(self, symbol: str, **kwargs):
    """
    Get trades for a specific account and symbol.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Account-Trade-List
    """
    endpoint = "/dapi/v1/userTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_position_risk(self, **kwargs):
    """
    Get current account information.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Position-Information
    """
    endpoint = "/dapi/v1/positionRisk"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_position_side_dual(
    self, dual_side_position: Literal["true", "false"], **kwargs
):
    """
    Change user's position mode (Hedge Mode or One-way Mode ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Change-Position-Mode#request-weight
    """
    endpoint = "/dapi/v1/positionSide/dual"
    payload = {"dualSidePosition": dual_side_position, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_margin_type(
    self, symbol: str, margin_type: Literal["ISOLATED", "CROSSED"], **kwargs
):
    """
    Change user's margin type in the specific symbol market.For Hedge Mode, LONG and SHORT positions of one symbol use the same margin type.
    With ISOLATED margin type, margins of the LONG and SHORT positions are isolated from each other.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Change-Margin-Type
    """
    endpoint = "/dapi/v1/marginType"
    payload = {"symbol": symbol, "marginType": margin_type, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_leverage(self, symbol: str, leverage: int, **kwargs):
    """
    Change user's initial leverage in the specific symbol market.
    For Hedge Mode, LONG and SHORT positions of one symbol use the same initial leverage and share a total notional value.
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Change-Initial-Leverage
    """
    endpoint = "/dapi/v1/leverage"
    payload = {"symbol": symbol, "leverage": leverage, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_adl_quantile(self, **kwargs):
    """
    Query position ADL quantile estimation
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Position-ADL-Quantile-Estimation
    """
    endpoint = "/dapi/v1/adlQuantile"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_dapi_v1_position_margin(
    self, symbol: str, amount: float, _type: Literal[1, 2], **kwargs
):
    """
    Modify Isolated Position Margin
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Modify-Isolated-Position-Margin
    """
    endpoint = "/dapi/v1/positionMargin"
    payload = {"symbol": symbol, "amount": amount, "type": _type, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_dapi_v1_position_margin_history(self, symbol: str, **kwargs):
    """
    Get position margin change history
    https://developers.binance.com/docs/derivatives/coin-margined-futures/trade/Get-Position-Margin-Change-History
    """
    endpoint = "/dapi/v1/positionMargin/history"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



from typing import Literal

from nexusbridge.binance.constants import Side


async def post_papi_v1_um_order(
    self, symbol: str, side: Side, _type: Literal["LIMIT", "MARKET"], **kwargs
):
    endpoint = "/papi/v1/um/order"
    """
    Place new UM order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade
    """
    payload = {"symbol": symbol, "side": side.value, "type": _type, **kwargs}
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def post_papi_v1_um_conditional_order(
    self,
    symbol: str,
    side: Side,
    strategy_type: Literal[
        "STOP",
        "STOP_MARKET",
        "TAKE_PROFIT",
        "TAKE_PROFIT_MARKET",
        "TRAILING_STOP_MARKET",
    ],
    **kwargs,
):
    endpoint = "/papi/v1/um/conditional/order"
    """
    Place new UM conditional order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/New-UM-Conditional-Order
    """
    payload = {
        "symbol": symbol,
        "side": side.value,
        "strategyType": strategy_type,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_papi_v1_cm_order(
    self, symbol: str, side: Side, _type: Literal["LIMIT", "MARKET"], **kwargs
):
    endpoint = "/papi/v1/cm/order"
    """
    Place new CM order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/New-CM-Order
    """
    payload = {"symbol": symbol, "side": side.value, "type": _type, **kwargs}
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def post_papi_v1_cm_conditional_order(
    self,
    symbol: str,
    side: Side,
    strategy_type: Literal[
        "STOP",
        "STOP_MARKET",
        "TAKE_PROFIT",
        "TAKE_PROFIT_MARKET",
        "TRAILING_STOP_MARKET",
    ],
    **kwargs,
):
    endpoint = "/papi/v1/cm/conditional/order"
    """
    New CM Conditional Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/New-CM-Conditional-Order
    """
    payload = {
        "symbol": symbol,
        "side": side.value,
        "strategyType": strategy_type,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_papi_v1_margin_order(
    self, symbol: str, side: Side, _type: Literal["LIMIT", "MARKET"], **kwargs
):
    endpoint = "/papi/v1/margin/order"
    """
    New Margin Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/New-Margin-Order
    """
    payload = {"symbol": symbol, "side": side.value, "type": _type, **kwargs}
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def post_papi_v1_margin_loan(self, asset: str, amount: float, **kwargs):
    endpoint = "/papi/v1/marginLoan"
    """
    Apply for a loan
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Margin-Account-Borrow
    """
    payload = {"asset": asset, "amount": amount, **kwargs}
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def post_papi_v1_repay_loan(self, asset: str, amount: float, **kwargs):
    endpoint = "/papi/v1/repayLoan"
    """
    Repay for a margin loan.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Margin-Account-Repay
    """
    payload = {"asset": asset, "amount": amount, **kwargs}
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def post_papi_v1_margin_order_oco(
    self,
    symbol: str,
    side: Side,
    quantity: float,
    price: float,
    stop_price: float,
    **kwargs,
):
    endpoint = "/papi/v1/margin/order/oco"
    """
    New Margin OCO Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Margin-Account-New-OCO
    """
    payload = {
        "symbol": symbol,
        "side": side.value,
        "quantity": quantity,
        "price": price,
        "stopPrice": stop_price,
        **kwargs,
    }
    raw = await self._fetch(
        "POST",
        endpoint,
        payload=payload,
        signed=True,
    )
    return raw


async def delete_papi_v1_um_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/order"
    """
    Cancel an active UM LIMIT order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-UM-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_um_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/allOpenOrders"
    """
    Cancel all active LIMIT orders on specific symbol
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-All-UM-Open-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_um_conditional_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/conditional/order"
    """
    Cancel UM Conditional Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-UM-Conditional-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_um_conditional_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/conditional/allOpenOrders"
    """
    Cancel All UM Open Conditional Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-All-UM-Open-Conditional-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_cm_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/order"
    """
    Cancel an active LIMIT order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-CM-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_cm_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/allOpenOrders"
    """
    Cancel all active LIMIT orders on specific symbol
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-All-CM-Open-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_cm_conditional_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/conditional/order"
    """
    Cancel CM Conditional Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-CM-Conditional-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_cm_conditional_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/conditional/allOpenOrders"
    """
    Cancel All CM Open Conditional Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-All-CM-Open-Conditional-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_margin_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/order"
    """
    Cancel Margin Account Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-Margin-Account-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_margin_orderList(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/orderList"
    """
    Cancel Margin Account OCO Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-Margin-Account-OCO-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_papi_v1_margin_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/allOpenOrders"
    """
    Cancel Margin Account All Open Orders on a Symbol
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Cancel-Margin-Account-All-Open-Orders-on-a-Symbol
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def put_papi_v1_um_order(
    self, side: Side, symbol: str, quantity: float, **kwargs
):
    endpoint = "/papi/v1/um/order"
    """
    Order modify function, currently only LIMIT order modification is supported, modified orders will be reordered in the match queue
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Modify-UM-Order
    """
    payload = {"symbol": symbol, "side": side.value, "quantity": quantity, **kwargs}
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def put_papi_v1_cm_order(
    self, side: Side, symbol: str, quantity: float, price: float, **kwargs
):
    endpoint = "/papi/v1/cm/order"
    """
    Order modify function, currently only LIMIT order modification is supported, modified orders will be reordered in the match queue
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Modify-CM-Order
    """
    payload = {
        "symbol": symbol,
        "side": side.value,
        "quantity": quantity,
        "price": price,
        **kwargs,
    }
    raw = await self._fetch("PUT", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/order"
    """
    Check an UM order's status.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-UM-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_all_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/allOrders"
    """
    Get all account UM orders; active, canceled, or filled.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-UM-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_open_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/openOrder"
    """
    Query current UM open order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Current-UM-Open-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_open_orders(self, **kwargs):
    endpoint = "/papi/v1/um/openOrders"
    """
    Get all open orders on a symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-Current-UM-Open-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_conditional_all_orders(self, **kwargs):
    endpoint = "/papi/v1/um/conditional/allOrders"
    """
    Query All UM Conditional Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-UM-Conditional-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_conditional_open_orders(self, **kwargs):
    endpoint = "/papi/v1/um/conditional/openOrders"
    """
    Get all open conditional orders on a symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-Current-UM-Open-Conditional-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_conditional_open_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/conditional/openOrder"
    """
    Query Current UM Open Conditional Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Current-UM-Open-Conditional-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_conditional_order_history(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/conditional/orderHistory"
    """
    Query UM Conditional Order History
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-UM-Conditional-Order-History
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/order"
    """
    Check an CM order's status.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-CM-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_all_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/allOrders"
    """
    Get all account CM orders; active, canceled, or filled.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-CM-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_open_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/openOrder"
    """
    Query current CM open order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Current-CM-Open-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_open_orders(self, **kwargs):
    endpoint = "/papi/v1/cm/openOrders"
    """
    Get all open orders on a symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-Current-CM-Open-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_conditional_open_orders(self, **kwargs):
    endpoint = "/papi/v1/cm/conditional/openOrders"
    """
    Get all open conditional orders on a symbol. Careful when accessing this with no symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-Current-CM-Open-Conditional-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_conditional_open_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/conditional/openOrder"
    """
    Query Current CM Open Conditional Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Current-CM-Open-Conditional-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_conditional_all_orders(self, **kwargs):
    endpoint = "/papi/v1/cm/conditional/allOrders"
    """
    Query All CM Conditional Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-CM-Conditional-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_conditional_order_history(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/conditional/orderHistory"
    """
    Query CM Conditional Order History
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-CM-Conditional-Order-History
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_force_orders(self, **kwargs):
    endpoint = "/papi/v1/um/forceOrders"
    """
    Query User's UM Force Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Users-UM-Force-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_force_orders(self, **kwargs):
    endpoint = "/papi/v1/cm/forceOrders"
    """
    Query User's CM Force Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Users-CM-Force-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_order_amendment(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/orderAmendment"
    """
    Get order modification history
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-UM-Modify-Order-History
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_order_amendment(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/cm/orderAmendment"
    """
    Get order modification history
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-CM-Modify-Order-History
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_force_orders(self, **kwargs):
    endpoint = "/papi/v1/margin/forceOrders"
    """
    Query user's margin force orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Users-Margin-Force-Orders
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_user_trades(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/um/userTrades"
    """
    Get trades for a specific account and UM symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/UM-Account-Trade-List
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_user_trades(self, **kwargs):
    endpoint = "/papi/v1/cm/userTrades"
    """
    Get trades for a specific account and CM symbol.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/CM-Account-Trade-List
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_adl_quantile(self, **kwargs):
    endpoint = "/papi/v1/um/adlQuantile"
    """
    Query UM Position ADL Quantile Estimation
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/UM-Position-ADL-Quantile-Estimation
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_cm_adl_quantile(self, **kwargs):
    endpoint = "/papi/v1/cm/adlQuantile"
    """
    Query CM Position ADL Quantile Estimation
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/CM-Position-ADL-Quantile-Estimation
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_papi_v1_um_feeBurn(self, fee_burn: Literal["true", "false"], **kwargs):
    endpoint = "/papi/v1/um/feeBurn"
    """
    Change user's BNB Fee Discount for UM Futures (Fee Discount On or Fee Discount Off ) on EVERY symbol
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Toggle-BNB-Burn-On-UM-Futures-Trade
    """
    payload = {"feeBurn": fee_burn, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_um_feeBurn(self, **kwargs):
    endpoint = "/papi/v1/um/feeBurn"
    """
    Get user's BNB Fee Discount for UM Futures (Fee Discount On or Fee Discount Off )
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Get-UM-Futures-BNB-Burn-Status
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_order(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/order"
    """
    Query Margin Account Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Margin-Account-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_openOrders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/openOrders"
    """
    Query Current Margin Open Order
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Current-Margin-Open-Order
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_all_open_orders(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/allOrders"
    """
    Query All Margin Account Orders
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-All-Margin-Account-Orders
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_orderList(self, **kwargs):
    endpoint = "/papi/v1/margin/orderList"
    """
    Retrieves a specific OCO based on provided optional parameters
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Margin-Account-OCO
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_all_orderList(self, **kwargs):
    endpoint = "/papi/v1/margin/allOrderList"
    """
    Query all OCO for a specific margin account based on provided optional parameters
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Margin-Account-all-OCO
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_openOrderList(self, **kwargs):
    endpoint = "/papi/v1/margin/openOrderList"
    """
    Query Margin Account's Open OCO
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Query-Margin-Account-Open-OCO
    """
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_papi_v1_margin_myTrades(self, symbol: str, **kwargs):
    endpoint = "/papi/v1/margin/myTrades"
    """
    Margin Account Trade List
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Margin-Account-Trade-List
    """
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_papi_v1_margin_repay_debt(self, asset: str, **kwargs):
    endpoint = "/papi/v1/margin/repay-debt"
    """
    Repay debt for a margin loan.
    https://developers.binance.com/docs/derivatives/portfolio-margin/trade/Margin-Account-Repay-Debt
    """
    payload = {"asset": asset, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw
